/** Collectable.cpp
 * 
 * See header file for details.
 */

#include "Collectable.h"

#include <algorithm>

Collectable::Collectable(const char *typeName, const char *atlasFileName) : physicsObject(this) {
    state = CS_AVAILABLE;

    textureAtlas = std::make_shared<TextureAtlas>(atlasFileName);

    sprite = textureAtlas->getSprite(typeName);
    if(!sprite) {
        throw std::runtime_error(TextFormat("Couldn't find sprite %s in texture atlas %s.", typeName, atlasFileName));
    }

    auto *defaultAnim = sprite->getAnimation("default");
    if(!defaultAnim) {
        throw std::runtime_error(TextFormat("Texture atlas %s doesn't have a \"default\" animation for sprite %s.", atlasFileName, typeName));
    }
    defaultAnimPlayer.reset(new TASpriteAnimationPlayer(textureAtlas, defaultAnim));

    auto *collectAnim = sprite->getAnimation("collect");
    if(collectAnim) {
        collectAnimPlayer.reset(new TASpriteAnimationPlayer(textureAtlas, collectAnim));
    }

    currAnimPlayer = defaultAnimPlayer.get();
}

Collectable::~Collectable() {
}

bool Collectable::update(float elapsedTime) {
    switch(state) {
    case CS_AVAILABLE:
        defaultAnimPlayer->update(elapsedTime);
        break;
    case CS_COLLECTING:
        defaultAnimPlayer->update(elapsedTime);
        break;
    case CS_COLLECTED:
    default:
        // Collected. No longer exists in the world (outside of the player).
        die();
    }
	return isAlive;
}

void Collectable::draw(float worldScale) {
    auto &b2Pos = getPosition();
    raylib::Vector2 position(b2Pos.x, b2Pos.y);
    switch(state) {
    case CS_AVAILABLE:
        defaultAnimPlayer->draw(position, worldScale);
        break;
    case CS_COLLECTING:
        defaultAnimPlayer->draw(position, worldScale);
        break;
    case CS_COLLECTED:
    default:
        // Collected. No longer exists in the world (outside of the player).
        die();
    }
}

void Collectable::createPhysicsBody(b2World &world, float worldScale, 
		const b2Vec2 &position, const b2Vec2 &velocity) {
    const raylib::Vector2 &size = sprite->size / worldScale;

    b2BodyDef bodyDef;
	bodyDef.type = b2_dynamicBody;
	bodyDef.fixedRotation = true;
	bodyDef.position = position;
	bodyDef.linearVelocity = velocity;
    physicsBody = world.CreateBody(&bodyDef);

    // Treating the collectable as circle
    float halfWidth = size.x / 2.0f;
	float halfHeight = size.y / 2.0f;
    float radius = std::min(halfWidth, halfHeight);

    b2CircleShape shape;
	shape.m_radius = radius;
	b2FixtureDef shapeDef;
	shapeDef.shape = &shape;
	shapeDef.density = 1.0f;
    physicsObject.attachToFixture(shapeDef);
	physicsBody->CreateFixture(&shapeDef);
}
